#include <limits.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <sys/shm.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <time.h>
#include <unistd.h>
#include "config.h"
#include "msgqueue.h"

#define MAXSIZE 4096

// methods for bit array
#define SetBit(A, k)    ( A[(k/32)] |= (1 << (k%32)) )
#define ClearBit(A, k)  ( A[(k/32)] &= ~(1 << (k%32)) )
#define TestBit(A, k)   ( A[(k/32)] & (1 << (k%32)) )

typedef struct {
    long mtype;
    char mtext[MAXSIZE];
} mymsg_t;

struct node {
   int lpid;
   struct node *next;
};

// globals
char* logfile;
struct node *head = NULL;
struct node *current = NULL;
int msgid;
int segment_id[3];
int* sysclock;
int* timespent;
pcb_t* pcbs;
int cpuidlesec = 0;
int cpuidlens = 0;
int iowaittotalsec = 0;
int iowaittotalns = 0;
int cpuwaittotalsec = 0;
int cpuwaittotalns = 0;
int ioproctotal = 0;
int cpuproctotal = 0;
int iocputotalns = 0;
int cpucputotalns = 0;
int iotimesincpu = 0;
int cputimesincpu = 0;
int iotimeinsyssec = 0;
int cputimeinsyssec = 0;
int iotimeinsysns = 0;
int cputimeinsysns = 0;
int iotimeinblocksec = 0;
int iotimeinblockns = 0;
int cputimeinblocksec = 0;
int cputimeinblockns = 0;

// prototypes
void killandfree();
void handle_sigint(int sig);
void logmsg(char* msg, int opt);
void insertpid(int lpid);
struct node* removepid(int lpid);
static void handle_timer(int sig);
static int setupinterrupt(void);
static int setuptimer(int time);

int main(int argc, char* argv[]) {
    signal(SIGINT, handle_sigint);
    signal(SIGCHLD, SIG_IGN);

    // setting error message
    char errmsgraw[100];
    snprintf(errmsgraw, sizeof(errmsgraw), "%s: Error", argv[0]);
    char* errmsg = errmsgraw+2;

    int c;
    int sflag, lflag;
    char* svalue = NULL;
    char* lvalue = NULL;
    char* usagestmt = "Usage: oss [-h] [-s t] [-l f]";

    // reading and validating arguments
    while ((c = getopt(argc, argv, "hs:l:")) != -1) {
        switch(c) {
            case 'h':
                printf("%s\n", usagestmt);
                exit(EXIT_SUCCESS);
            case 's':
                sflag = 1;
                svalue = optarg;
                break;
            case 'l':
                lflag = 1;
                lvalue = optarg;
                break;
            default:
                exit(EXIT_FAILURE);
        }
    }
    
    // storing arguments
    int time;
    
    if (sflag > 0)
        time = atoi(svalue);
    else
        time = TIME_LIMIT;
    
    if (lflag > 0)
        logfile = lvalue;
    else
        logfile = "logfile";

    
    // setting up timer
    if (setupinterrupt() == -1) {
        snprintf(errmsgraw, sizeof(errmsgraw),
                errmsg, ": Failed to setup handler for SIGPROF");
        char* setuperr = errmsgraw;
        perror(setuperr);
        exit(EXIT_FAILURE);
    }
    if (setuptimer(time) == -1) {
        snprintf(errmsgraw, sizeof(errmsgraw),
                errmsg, ": Failed to setup ITIMER_PROF interval timer");
        char* setuperr = errmsgraw;
        perror(setuperr);
        exit(EXIT_FAILURE);
    }

    // allocating shared memory for system clock
    segment_id[0] = shmget(IPC_PRIVATE, 2*sizeof(int),
            IPC_CREAT | IPC_EXCL | S_IRUSR | S_IWUSR);
    sysclock = (int*) shmat(segment_id[0], NULL, 0);

    // allocating shared memory for time spent by a process
    segment_id[1] = shmget(IPC_PRIVATE, sizeof(int),
            IPC_CREAT | IPC_EXCL | S_IRUSR | S_IWUSR);
    timespent = (int*) shmat(segment_id[1], NULL, 0);

    // sysclock[0] holds seconds and sysclock[1] holds nanoseconds
    sysclock[0] = 0;
    sysclock[1] = 0;

    // allocating shared memory for process control blocks
    segment_id[2] = shmget(IPC_PRIVATE, 18*sizeof(pcb_t),
            IPC_CREAT | IPC_EXCL | S_IRUSR | S_IWUSR);
    pcbs = (pcb_t*) shmat(segment_id[2], NULL, 0);
   
    // initialize msg queue
    key_t key;    
    key = ftok("oss", 65);
    msgid = msgget(key, 0666 | IPC_CREAT);

    // turning segment ids to strings
    char buff[50];
    snprintf(buff, 49, "%d", segment_id[2]);
    char* pcbs_id_str = buff;
    char clockbuff[50];
    snprintf(clockbuff, 49, "%d", segment_id[0]);
    char* clock_id_str = clockbuff;

    // initializing bit array
    long A[1];
    A[0] = 0;

    int i = 0;
    mymsg_t mymsg;
    int highflag = 1;
    int whichqueue = 0;
    int makebabyns = 0;
    int makebabysec = 0;
    int numoflines = 0;
    int ioinittimesec[18];
    int ioinittimens[18];
    int ioblockedtimesec[18];
    int ioblockedtimens[18];
    int cpuinittimesec[18];
    int cpuinittimens[18];
    int cpublockedtimesec[18];
    int cpublockedtimens[18];
    int timetoswitch = 4;


    // initialize queues
    struct Queue* highqueue = createQueue(1000);
    struct Queue* lowqueue = createQueue(1000);
    
    logmsg("OSS: Starting Simulation...", 1);
    numoflines++;

    // seeding random nums
    srand(getpid());
    
    // oss main loop
    for (;;) {

        // check to see if any process is ready to be unblocked
        if (head != NULL) {
            struct node* tmp = head;
            while(tmp->next != NULL) {
                if (pcbs[tmp->lpid].unblocksec <= sysclock[0] || pcbs[tmp->lpid].unblockns <= sysclock[1]) {
                    enqueue(highqueue, tmp->lpid);
                    struct node* dlt = removepid(tmp->lpid);
                    if (pcbs[tmp->lpid].isIO) {
                        iotimeinblocksec = sysclock[0] - ioblockedtimesec[tmp->lpid];
                        iotimeinblockns = sysclock[1] - ioblockedtimens[tmp->lpid];
                    }
                    else {
                        cputimeinblocksec = sysclock[0] - cpublockedtimesec[tmp->lpid];
                        cputimeinblockns = sysclock[1] - cpublockedtimens[tmp->lpid];
                    }
                    free(dlt);
                    //if a need to move allocate some more time to clock
                    int randinc = rand() % 1000;
                    sysclock[1] += randinc;
                    
                }
            }
            // checking final element in list
            if (pcbs[tmp->lpid].unblocksec <= sysclock[0] || pcbs[tmp->lpid].unblockns <= sysclock[1]) {
                enqueue(highqueue, tmp->lpid);
                struct node* dlt = removepid(tmp->lpid);
                free(dlt);
                //if a need to move allocate some more time to clock
                int randinc = rand() % 1000;
                sysclock[1] += randinc;
            }
        }
        
        // make child if it's the appropriate time
        if ((sysclock[0] > makebabysec 
               || (sysclock[0] == makebabysec && sysclock[1] >= makebabyns))
               && (ioproctotal + cpuproctotal) < MAX_PROCESSES) {
            // find an open spot
            for (i = 0; i < 18; i++) {
                if (TestBit(A, i) == 0)
                    break;
            }
            // record if full
            if (i == 18) {
                char buff2[50];
                snprintf(buff2, 49, "OSS: Process Table is Full.");
                char* pcbsfullstr = buff2;
                logmsg(pcbsfullstr, 0);
                numoflines++;
            }
            else {
                // decide if IO or CPU bound
                int isIO = rand() % 2;
                if (isIO) {
                    ioproctotal++;
                    ioinittimesec[i] = sysclock[0];
                    ioinittimens[i] = sysclock[1];
                }
                else {
                    cpuproctotal++;
                    cpuinittimesec[i] = sysclock[0];
                    cpuinittimens[i] = sysclock[1];
                }

                // set up proc ctrl block
                pcbs[i].isIO = isIO;
                pcbs[i].msgqueueid = msgid;
                pcbs[i].nanoseconds = 0;
                pcbs[i].seconds = 0;
                pcbs[i].msgtype = i+2; 
                if (isIO)
                    pcbs[i].timeslice = 2000;
                else
                    pcbs[i].timeslice = 5000;

                // fork the child
                pid_t pid = fork();
                if (pid == 0) {
                    char buff1[50];
                    snprintf(buff1, 49, "%d", i);
                    char* lpid_str = buff1;
                    execlp("./user", "./user", pcbs_id_str, lpid_str, clock_id_str, NULL);
                }
                if (pid < 0) {
                    snprintf(errmsgraw, sizeof(errmsgraw),
                            errmsg, ": Failed to fork child");
                    char* setuperr = errmsgraw;
                    perror(setuperr);
                    // terminate children
                    for (int k = 0; k < 18; k++)
                        if (TestBit(A, k))
                            kill(pcbs[k].pid, SIGINT);
                    killandfree();
                }
                SetBit(A, i);
                // log child was created
                char bufferkid[100];
                snprintf(bufferkid, 99, "OSS: Created child with PID %d at time %d:%d", 
                        i, sysclock[0], sysclock[1]);
                char* message = bufferkid;
                logmsg(message, 0);
                numoflines++;
                
                // put process in appropriate queue
                if (isIO)
                    enqueue(highqueue, i);
                else
                    enqueue(lowqueue, i);
            } 
            // assign next time to make a child
            makebabysec = (rand() % MAXTIMENEWPROCSECS) + sysclock[0];
            makebabyns = (rand() % MAXTIMENEWPROCNS) + sysclock[1];
        }

        // if max processes, check that 
        else if ((ioproctotal + cpuproctotal) == 50) {
            int l;
            for (l = 0; l < 18; l++)
                if(TestBit(A, l) == 1)
                    break;
            if (l == 18)
                killandfree();
        }
        
        // get next process in ready queue
        if (sysclock[0] >= timetoswitch) {
            highflag = 1 - highflag;
            timetoswitch += 4;
        }
        int lpid;
        if (highflag)
            lpid = dequeue(highqueue);
        else
            lpid = dequeue(lowqueue);
        whichqueue = 1 - highflag;

        // increment clock for this iteration
        int randns = (rand() % 1000);
        sysclock[0] += 1;
        sysclock[1] += randns;
        while (sysclock[1] > 999999999) {
            sysclock[1] -= 1000000000;
            sysclock[0] += 1;
        }
        
        // if there's no processes in one queue, try the other
        if (lpid == INT_MIN) {
            if (highflag) {
                lpid = dequeue(lowqueue);
                whichqueue = 1;
            }
            else {
                lpid = dequeue(highqueue);
                whichqueue = 0;
            }
        }
        // if there's no procs in either queue, start next loop
        if (lpid == INT_MIN) {
            cpuidlesec += 1;
            cpuidlens += randns;
            continue;
        }
        
        // log who you're dispatching
        char buffer[100];
        snprintf(buffer, 99, "OSS: Dispatching PID %d from queue %d at time %d:%d", 
                lpid, whichqueue, sysclock[0], sysclock[1]);
        char* message = buffer;
        logmsg(message, 0);
        numoflines++;
        
        // log time spent for this iteration
        snprintf(buffer, 99, "OSS: total time this dispatch was 1:%d", randns);
        message = buffer;
        logmsg(message, 0);
        numoflines++;

        // send message to process
        mymsg.mtype = lpid + 2;
        strcpy(mymsg.mtext, "Wake up, dummy");
        msgsnd(msgid, &mymsg, sizeof(mymsg), 0);
        
        // record childs wait time
        if (pcbs[lpid].isIO) {
            iowaittotalsec += (sysclock[0] - pcbs[lpid].waitstartsec);
            iowaittotalns += (sysclock[1] - pcbs[lpid].waitstartns);
        }
        else {
            cpuwaittotalsec += (sysclock[0] - pcbs[lpid].waitstartsec);
            cpuwaittotalns += (sysclock[1] - pcbs[lpid].waitstartns);
        }

        // wait for message from child
        msgrcv(msgid, &mymsg, sizeof(mymsg), 22, 0);
        
        // record childs cpu utilization
        if (pcbs[lpid].isIO) {
            iocputotalns += pcbs[lpid].nanoseconds;
            iotimesincpu++;
        }
        else {
            cpucputotalns += pcbs[lpid].nanoseconds;
            cputimesincpu++;
        }

        // log childs response and increment clock
        snprintf(buffer, 99, "OSS: Receiving that PID %d ran for %d nanoseconds",
                lpid, pcbs[lpid].nanoseconds);
        message = buffer;
        logmsg(message, 0);
        numoflines++;
        sysclock[1] += pcbs[lpid].nanoseconds;
        while (sysclock[1] > 999999999) {
            sysclock[1] -= 1000000000;
            sysclock[0] += 1;
        }

        // log if the child blocked
        if (mymsg.mtext[0] == 'B') {
            insertpid(lpid);
            snprintf(buffer, 99, "OSS: %d is blocked. Moving to blocked queue", lpid);
            message = buffer;
            logmsg(message, 0);
            numoflines++;
            // store blocked time
            if (pcbs[lpid].isIO) {
                ioblockedtimesec[lpid] = sysclock[0];
                ioblockedtimens[lpid] = sysclock[1];
            }
            else {
                cpublockedtimesec[lpid] = sysclock[0];
                cpublockedtimens[lpid] = sysclock[1];
            }
        }

        // log if the child terminated
        else if (mymsg.mtext[0] == 'T') {
            snprintf(buffer, 99, "OSS: %d has terminated at time %d:%d", 
                    lpid, sysclock[0], sysclock[1]);
            message = buffer;
            logmsg(message, 0);
            numoflines++;
            // store time in system
            if (pcbs[lpid].isIO) {
                iotimeinsyssec += (sysclock[0] - ioinittimesec[lpid]);
                iotimeinsysns += (sysclock[1] - ioinittimens[lpid]);
            }
            else {
                cputimeinsyssec += (sysclock[0] - cpuinittimesec[lpid]);
                cputimeinsysns += (sysclock[1] - cpuinittimens[lpid]);
            }
            ClearBit(A, lpid);
        }

        // if neither, then put back in ready queue
        else {
            if (pcbs[lpid].isIO)
                enqueue(highqueue, lpid);
            else
                enqueue(lowqueue, lpid);
        }

        // check if at MAXLINES
        if (numoflines >= MAXLINES) {
            snprintf(buffer, 99, "OSS: Total lines acheived.");
            message = buffer;
            logmsg(message, 0);
            
            // kill all child processes
            for (int k = 0; k < 18; k++)
                if (TestBit(A, k))
                    kill(pcbs[k].pid, SIGINT);
            
            //clear memory
            killandfree();
        }
    }
    
    // clear memeory
    killandfree();
}


// cleans up shared mem, writes summary to log
// and kills the message queue
void killandfree() {
    shmdt(sysclock);
    shmdt(timespent);
    shmctl(segment_id[0], IPC_RMID, 0);
    shmctl(segment_id[1], IPC_RMID, 0);
    shmctl(segment_id[2], IPC_RMID, 0);
    
    logmsg("OSS: Ending Simulation...", 0);

    // logging summary
    logmsg("\nSUMMARY", 0);
    char buffer[100];

    // idle cpu time
    snprintf(buffer, 99, "Idle CPU time: %d:%d", cpuidlesec, cpuidlens);
    char* message = buffer;
    logmsg(message, 0);
    
    // avg wait time
    int avgsec = cpuwaittotalsec / cpuproctotal;
    int avgns = cpuwaittotalns / cpuproctotal;
    while (avgns > 999999999) {
        avgns -= 1000000000;
        avgsec++;
    }
    snprintf(buffer, 99, "Average CPU-Bound Process wait time: %d:%d", avgsec, avgns);
    message = buffer;
    logmsg(message, 0);
    avgsec = iowaittotalsec / ioproctotal;
    avgns = iowaittotalns / ioproctotal;
    while (avgns > 999999999) {
        avgns -= 1000000000;
        avgsec++;
    }
    snprintf(buffer, 99, "Average IO-Bound Process wait time: %d:%d", avgsec, avgns);
    message = buffer;
    logmsg(message, 0);

    // avg cpu utilization
    if (cputimesincpu == 0)
        avgns = 0;
    else
        avgns = cpucputotalns / cputimesincpu;
    snprintf(buffer, 99, "Average CPU-Bound Process cpu utilization: %d nanoseconds", avgns);
    message = buffer;
    logmsg(message, 0);
    avgns = iocputotalns / iotimesincpu;
    snprintf(buffer, 99, "Average IO-Bound Process cpu utilization: %d nanoseconds", avgns);
    message = buffer;
    logmsg(message, 0);
    
    // avg time in sys
    avgsec = cputimeinsyssec / cpuproctotal;
    avgns = cputimeinsysns / cpuproctotal;
    while (avgns > 999999999) {
        avgns -= 1000000000;
        avgsec++;
    }
    snprintf(buffer, 99, "Average CPU-Bound Process time in system: %d:%d", avgsec, avgns);
    message = buffer;
    logmsg(message, 0);
    avgsec = iotimeinsyssec / ioproctotal;
    avgns = iotimeinsysns / ioproctotal;
    while (avgns > 999999999) {
        avgns -= 1000000000;
        avgsec++;
    }
    snprintf(buffer, 99, "Average IO-Bound Process time in system: %d:%d", avgsec, avgns);
    message = buffer;
    logmsg(message, 0);

    // avg time in blocked queue
    avgsec = cputimeinblocksec / cpuproctotal;
    avgns = cputimeinsysns / cpuproctotal;
    while (avgns > 999999999) {
        avgns -= 1000000000;
        avgsec++;
    }
    snprintf(buffer, 99, "Average CPU-Bound Process time in block queue: %d:%d", avgsec, avgns);
    message = buffer;
    logmsg(message, 0);
    avgsec = iotimeinblocksec / ioproctotal;
    avgns = iotimeinsysns / ioproctotal;
    while (avgns > 999999999) {
        avgns -= 1000000000;
        avgsec++;
    }
    snprintf(buffer, 99, "Average IO-Bound Process time in block queue: %d:%d", avgsec, avgns);
    message = buffer;
    logmsg(message, 0);

    // killing msg queue and exitting
    msgctl(msgid, IPC_RMID, NULL);
    exit(EXIT_SUCCESS);
}

// called when SIGINT is caught
void handle_sigint(int sig) {
    printf("\nSIGINT caught!\n");
    killandfree();
}

// used to log messages to a specified file
void logmsg(char* msg, int opt) {
    if (opt != 1) {
        FILE *fptr;
        fptr = fopen(logfile, "a");
        fprintf(fptr, "%s\n", msg);
        fclose(fptr);
    }
    else {
        FILE *fptr;
        fptr = fopen(logfile, "w");
        fprintf(fptr, "%s\n", msg);
        fclose(fptr);
    }
}

//insert node
void insertpid(int lpid) {
   //create a link
   struct node *link = (struct node*) malloc(sizeof(struct node));
   link->lpid = lpid;
   link->next = head;
   head = link;
}

//delete node with given pid
struct node* removepid(int lpid) {
   struct node* current = head;
   struct node* previous = NULL;
   if(head == NULL) {
      return NULL;
   }
   while(current->lpid != lpid) {
      if(current->next == NULL) {
         return NULL;
      } else {
         previous = current;
         current = current->next;
      }
   }
   if(current == head) {
      head = head->next;
   } else {
      previous->next = current->next;
   }    
   return current;
}

// called when the timer interval has occurred
void handle_timer(int sig) {
    printf("Times up! Killing children and freeing shared memory.\n");
    killandfree();
}

// sets up the interrupt for timer interval
static int setupinterrupt(void) {
    struct sigaction act;
    act.sa_handler = handle_timer;
    act.sa_flags = 0;
    return (sigemptyset(&act.sa_mask) || sigaction(SIGALRM, &act, NULL));
}

// sets up timer
static int setuptimer(int time) {
    struct itimerval value;
    value.it_interval.tv_sec = time;
    value.it_interval.tv_usec = 0;
    value.it_value = value.it_interval;
    return (setitimer(ITIMER_REAL, &value, NULL));
}

